/* $Id$
 *
 *  UNIX-Connect, a ZCONNECT(r) Transport and Gateway/Relay.
 *  Copyright (C) 1992-94  Martin Husemann
 *  Copyright (C) 1995     Christopher Creutzig
 *  Copyright (C) 1994-98  Moritz Both
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 * Eine deutsche Zusammenfassung zum Copyright dieser Programme sowie der
 * GNU General Public License finden Sie in der Datei README.1st.
 * ------------------------------------------------------------------------
 *
 *  Bugreports, suggestions for improvement, patches, ports to other systems
 *  etc. are welcome. Contact the maintainer by e-mail:
 *  christopher@nescio.foebud.org or snail-mail:
 *  Christopher Creutzig, Im Samtfelde 19, 33098 Paderborn
 *
 *  There is a mailing-list for user-support:
 *   unix-connect@mailinglisten.im-netz.de,
 *  write a mail with subject "Help" to
 *   unix-connect-request@mailinglisten.im-netz.de
 *  for instructions on how to join this list.
 */


/*
 * uuconv.c - Funktionen fuer Konvwertierung von RFC nach 
 * ZConnect, die fuer Mail und News (fast) gleich sind
 *
 * $Log$
 * Revision 1.29  2000/02/24 15:31:45  moritz
 * nderungen aus dem Projekt KeinZerberus2000 eingefgt
 *
 * Revision 1.28  1998/04/17 16:02:38  moritz
 * - bugfix mime in Hd-Zeilen
 *
 * Revision 1.27  1998/04/17 12:54:08  moritz
 * - Compilierungs-Zhler in gate/
 * - References: werden nicht mehr umgebrochen
 * - umbannant in UnixConnect II
 * - einige Bugfixes von der offiziellen Liste bernommen
 * - NOKOP wird korrekt behandelt
 * - base64-dekodierung von Headerzeilen, base64-dekodierung allg. verbessert
 *
 * Revision 1.26  1998/04/16 14:24:43  moritz
 * - Offizielle Version 0.76b1 hinzugefgt
 * - Erzeugt Message-ID, wenn keine da war bei Mail
 * - qp-De- und Enkodierung auch von X-Z- und X-ZC- Headerzeilen
 * - NOKOP wird korrekt behandelt: Keine Erzeugung von
 *   verrterischen To: oder Cc: Headerzeilen, wenn es
 *   vorhanden ist
 *
 * Revision 1.25  1998/03/04 02:03:57  moritz
 * - fix ROT/Path: handling
 *
 * Revision 1.24  1998/03/04 01:51:46  moritz
 * - code von mime.c in neues file mimew.c verschoben (nur fr Wandlung
 *         ZC->RFC)
 * - funktion 'count' in library verschoben
 * - Problem mit zuvielen/zuwenig Leerzeichen am Nachrichtenanfang behoben
 * - Q-Codierung von Headerzeilen weiter RFC 2047 angenhert (Zeilenlnge,
 *         Unterstriche)
 *
 * Revision 1.23  1998/02/17 11:52:04  moritz
 * - bugfix
 *
 * Revision 1.22  1998/02/17 09:27:29  moritz
 * - konvertiere 8-Bit-codierte Headerzeilen RFC->ZC
 *
 * Revision 1.21  1998/02/13 13:23:29  moritz
 * - split make targets 'mail' and 'news'
 * - added rcs Ids into main programs
 *
 * Revision 1.20  1998/02/04 22:44:23  moritz
 * - besserer Headerzeilen-Umbruch bei Wandlung ZC->RFC
 *
 * Revision 1.19  1998/02/03 22:34:58  moritz
 * - kein Path: bei Mail
 * - Adressen 'xxx@uucp.do.main' werden nicht verstuemmelt (Danke, Matthias)
 * - Filenamen-Erzeugung der nach-ZConnect-Konverter: Sie heissen jetzt erst
 *     dann *.prv oder *.brt, wenn sie fertig sind
 *
 * Revision 1.18  1998/02/03 22:00:11  moritz
 * - compilierende, ungetestete Veriosn von 0.76MB01
 *
 * Revision 1.17  1998/02/03 16:18:39  moritz
 * - merged diffs from 0.75b2 to 0.76 into MB Version
 *
 * Revision 1.16  1998/02/02 16:29:25  moritz
 * - Aenderungen der offiziellen Version bis 0.75b2 eingearbeitet
 *
 * Revision 1.15  1996/08/15 19:32:25  moritz
 * Bugfix From: mit mehreren @ nach ABS: Bisher wurden
 * defekte ZConnect- Nach4richten ohne ABS: erzeugt
 *
 * Revision 1.14  1996/08/12 06:38:04  moritz
 * - 0.74b4 integriert (u.a. Dateinamen-Erzuegung verbessert)
 *
 * Revision 1.13  1996/07/25 09:42:50  moritz
 * APC-atob rausgeworfen!
 * Dadurch hoffentlich core bei sehr grossen Mails verhindert.
 *
 * Revision 1.12  1996/07/16 10:12:00  moritz
 * Fixed bug in splittadr: would destroy addresses
 * with no space between addr and comment/real name
 *
*/

 
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef HAS_STRING_H
# include <string.h>
#endif
#ifdef HAS_STRINGS_H
# include <strings.h>
#endif
#ifdef NEED_VALUES_H
#include <values.h>
#endif
#ifdef HAS_UNISTD_H
# include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include "utility.h"
#include "header.h"
#include "hd_def.h"
#include "hd_nam.h"
#include "boxstat.h"
#include "version.h"
#include "uulog.h"
#include "alias.h"
#include "mime.h"
#ifdef APC_A2B
#  include "apc_a2b.h"
#endif
#include "lib.h"
#include "cnumber.h"

extern char *fqdn;	/* FQDN des Hosts, fuer den hier ZCONNECT erzeugt wird */
extern char *netcalldir;
char wab_name[MAXLINE] = "", wab_host[MAXLINE] = "", wab_domain[MAXLINE] = "";
static char sp_name[MAXLINE], sp_host[MAXLINE], sp_domain[MAXLINE], rna[MAXLINE];
static char iso_8859_1[] = "iso-8859-1";
static char us_ascii[] =   "us-ascii";

time_t parsedate(char *date_header, int *time_zone);
static char *decode_mime_header(register char *s);

int dont_gate = 0;	/* Flag fuer Pseudo Z3.8 Nachrichten */

/* Open a unique file in the spool directory. Give it a temporary name,
 * so other programs wont touch it until closed and renamed.
 *
 * Return malloc'ed string containing the file name. Put the file handle
 * info *pfh. Return NULL if any error.
 */
char *open_unique_spoolfile(FILE **pfh) {
		time_t j;
                char *p, *datei; 
                int fh;
                struct stat sb;
#ifdef SPOOLDIR_SHORTNAME
                char *p1;
#endif
                
                datei = (char *)malloc(strlen(netcalldir)+strlen(fqdn)+30);
                if (!datei) {
                        fprintf(stderr, "open_unique_spoolfile: Out of memory\n");
                        exit(1); 
                }        
 
                strcpy(datei, netcalldir);
                strcat(datei, "/");
                strcat(datei, fqdn);
#ifdef SPOOLDIR_SHORTNAME
                p = strrchr(datei, '/');
                if (p) {
                        p1 = strchr(p+1, '.');
                        if (p1) *p1 = '\0';  
                }
#endif
                strcat(datei, "/");
                if (stat(datei, &sb) != 0 && errno==ENOENT) {
                        if (mkdir(datei, 02775) != 0) {
                                fprintf(stderr, "open_unique_spoolfile: mkdir(%s)
 failed: %s\n",
                                        datei, sys_errlist[errno]);
                                exit(1);
                        }
                }
		p = strchr(datei, 0); /* Ende des String merken */
                j = time(NULL);
		while(1) {
			sprintf(p, "%08lx", (long)j++);
			fh = open(datei, O_WRONLY|O_CREAT|O_EXCL, 0600);
			if (fh>=0 || errno != EEXIST)
				break;
		}
		if (fh<0) {
			dfree(datei);
			return NULL;
		}
		*pfh = fdopen(fh, "ab");
		return datei;
}

/* rename a spool file to a unique name with suffix suffix.
 * Return 1 on success, 0 otherwise.
 */
int rename_spoolfile(char *file, char *suffix) {
	char *p, *newname;
	time_t j;

	newname=(char *)dalloc(strlen(file)+5);
	strcpy(newname, file);
	p = strrchr(newname, '/');
	if (!p)
		p = newname;
	else
		p++;
	j = time(NULL);
	while(1) {
		sprintf(p, "%08lx.%s", (long)j++, suffix);
		if (link(file, newname) == 0)
			break;
		if (errno != EEXIST) {
			dfree(newname);
			return 0;
		}
	}
	unlink(file);
	return 1;
}

/* Hier werden die nach ZConnect gewandelten Brettnamen
   auf Gltigkeit geprft. */
static int valid_newsgroups( char *data )
{
	if(NULL==data)
		return 0;
	/* 0 zurueckgeben, wenn dies eine Mail-Adresse sein koennte */
	if ( strchr( data, '@' ) != NULL )
		return 0;
	/* Wir wollen Bretter mit '!' haben! */
	if ( *data != '/' )
		return 0;
	return 1;
}

/* Druckt alle angegebenen newsgroups (mit Komma unterteilt)
 * in eigene ZConnect-Headerzeilen.
 * Wenn zc_header NULL ist, druckt es nur eine Zeile aus
 * der ersten angegebenen newsgroup und fertig.
 */
static void printbretter(char *newsgroups, char *zc_header, FILE *f)
{
	char *adr;
	int adr_len;
	char *bak, *st, *q, *s;
	size_t len;

	/* get a buffer */
	adr_len=MAXLINE;
	adr = (char *) dalloc(adr_len);

	/* bearbeite jede einzelne Newsgrous einzeln */
	bak = dstrdup( newsgroups );
	st = strtok( bak, " , \t" );
	while ( st != NULL ) {
		do {
			/* vorher ".ctl" im Brettnamen abscheiden */
			len = strlen( st ) -4;
			if ( len > 0 ) {
				if ( strncmp( st + len, ".ctl", 4 ) == 0 )
					st[ len ] = '\0';
			};
			s = u_alias(st);
			if (s) {
				if (zc_header)
					fprintf(f, "%s: ", zc_header);
				fprintf(f, "%s\r\n", s);
				break;
			};
			s = u_prefix(st);
			len=0;
			if (s) {
				strncpy(adr, s, MAXLINE);
				adr[MAXLINE-1] = '\0';
				len=strlen(adr);
				st += prefix_len;
			} else {
				prefix_len = 0;
				adr[0]='/';
				len=1;
			};
			s=adr+len;
			/* Konvertiere diesen Gruppennamen */
			for (q=st; *q ; q++, len++ ) {
				if (len >= adr_len) {
					char *p;
					adr_len += 50;
					p=dalloc(adr_len);
					strcpy(p, adr);
					s = p + len;
					dfree(adr);
					adr=p;
				}
				if (*q == '.')
					*(s++) = '/';
				else
					*(s++) = toupper(*q);
			};
			*s = '\0';
			s = adr;
			if ( valid_newsgroups( adr ) != 0 )
			{
				if (zc_header)
					fprintf(f, "%s: ", zc_header);
				fprintf(f, "%s\r\n", s);
				break;
			};
#ifdef LOG_ERRORS_IN_HEADERS
			/* Sollte man einen Header generieren? */
			if (zc_header) {
				fprintf(f, "X-Bogus-Newsgroups-%s: ",
					zc_header);
				fprintf(f, "%s\r\n", st - prefix_len );
			}
#endif
			break;
		} while ( 0 );
		st = strtok( NULL, " , \t" );
	};
	dfree( bak );
	dfree( adr );
}

/*
 *   Findet das letzte '@' in einem full qualified domain name und liefert
 *     NULL -> kein @ enthalten
 *     != 0 -> Zeiger auf das '@'
 *   Hinter dem FQDN darf noch (Real Name) stehen!
 */
char *fqdn_at(char *s)
{
	char *erg;

	for (erg=NULL; *s && !isspace(*s); s++) 
		if (*s == '@')
			erg = s;
	return erg;
}

/*
 *   Findet das Komma, das diese Adresse von der naechsten Adresse
 *   trennt (z.B. in "To: x@y, a@b, c@d").
 *   Minimal (aber anscheinend hinreichend) wird RFC822 Syntax geparsed:
 *   Kommata in ( ) und in " " werden ignoriert.
 *
 *   Liefert:
 *     NULL  -> kein Komma, dies ist die letzte Adresse
 *     != 0  -> Zeiger auf das Komma
 */
char *next_komma(char *s)
{
	int level;
	
	while ( *s && *s != ',' ) {
		if (*s == '"') {
			s++;
			while(*s && *s != '"')
				s++;
			if (!*s) break;
		}
		if (*s == '(') {
			s++;
			level = 1;
			while ( *s && level ) {
				if (*s == '(')
					level++;
				else if (*s == ')')
					level--;
				s++;
			}
		} else {
			s++;
		}
	}

	if (*s)
		return s;
	else
		return NULL;
}


void splitaddr(char *rfc_addr, char *name, char *host, char *domain,
	char *rna)
{
	static char adr[MAXLINE];
	char *q, *q1, *s;

	/* "Foo <blah@ee.fds>" und
	 * "fds@fds.fds (Blah)"
	 * vereinheitlichen
	 */

	*name = '\0';
	*host = '\0';
	*domain = '\0';
	*rna = '\0';
	q = strchr(rfc_addr, '<');
	if (q) {
		strcpy(adr, q+1);
		q = strchr(adr, '>');
		if (q) *q = '\0';
		strcpy(rna, rfc_addr);
		q = strchr(rna, '<');
		while (isspace(*(--q) ))
			;
		q++;
		*q = '\0';
	} else {
		strcpy(adr, rfc_addr);
		/*
		 * Das hier ist noch nicht ganz korrekt. Laut RFC822
		 * ist 
		 *   Muhammed (I am the greatest!) .Ali @ somewhere.edu (Muhammed Ali)
		 * eine korrekte Adresse, die auf das Postfach
		 *   Muhammed.Ali@somewhere.edu
		 * zeigt. Es ist aber schon besser als die Lsung mit isspace,
		 * da Klammern, die nicht vom Adreteil getrennt 
		 * sind, erkannt werden.
		 */
		q=adr+strcspn(adr," \t(");
		/* for (q=adr; *q && !isspace(*q); q++)
			; */
		s = NULL;
		if(*q) 
			if ((s = strchr(q, '(')) != NULL)
				s++;
		*q = '\0';
		if (s) {
			strcpy(rna, s);
			s = strrchr(rna, ')');
			if (s) *s = '\0';
		}
		*q='\0';
	}

	/*
	 *   : Notation in ! wandeln
	 */
	
	for (q=adr; *q; q++)
		if (*q == ':') *q = '!';
	q = fqdn_at(adr);
	if (q) {
		int dots, pre_dots;
		
		*q = '\0'; pre_dots = dots = 0;
		for (s=adr; *s; s++) {
			if (*s == '.') {
				dots++;
			} else if (*s == '!') {
				pre_dots = dots;
				dots = 0;
			}
		}
		if (pre_dots && !dots) 
			q = NULL;
		else
			*q = '@';
	}
	s = strchr(adr, '!');
	if (s && !q) {
		/* Reine Bang-Adresse */
		for (q=q1=NULL,s=adr; *s; s++)
			if (*s == '!') {
				if (q)
					q1 = q+1;
				q = s;
			} else if (isspace(*s))
				break;
		*s = '\0';
		if (q) *q = '\0';
		if (!q1) q1 = adr;
		if (q) {
			strcpy(name, q+1);
			strcpy(host, q1);
			q = strchr(host, '.');
			if (q) {
				*q = '\0';
				strcpy(domain, q+1);
			} else {
				domain[0] = '\0';
			}
		} else {
			host[0] = '\0';
			domain[0] = '\0';
			strcpy(name, adr);
		}
	} else if (q) {
		/* Internet-Adresse */
		*q = '\0';
		strcpy(name, adr);
		strcpy(host, q+1);
		q = strchr(host, '.');
		if (q) {
			*q = '\0';
			strcpy(domain, q+1);
		} else {
			domain[0] = '\0';
		}
	}
	 else if (!s && !q) {
	 /* Die Adresse enthlt nur den Usernamen.
	  * Das ist zwar vollkommen illegal, mehrdeutig
	  * und unsinnig, aber das Beste, was wir tun
	  * koennen, ist wohl, den Namen einfach durchzureichen.
	  */
	  	host[0] = '\0';
	  	domain[0] = '\0';
	  	strcpy(name, adr);
	  }
}

/*
 *	max == 0 beliebig viele Adressen
 *	max > 0  maximal max Adressen. (es werden die ersten max ausgegeben)
 *
 *   return:
 *		die Anzahl der ausgegebenen Adressen (alle ungltigen werden
 *		nicht erzeugt)
 */
int convaddr(char *zconnect_header, char *rfc_addr, int max, FILE *f)
{
	char *komma, *start, *line;
	static char zaddr[MAXLINE];
	int counter;
	char *q;

	counter = 0;
	line = dstrdup(rfc_addr);
	for (start = line; start && *start ; ) {
		while (isspace(*start))
			start++;
		komma = next_komma(start);
		if (komma) *komma = '\0';
		splitaddr(start, sp_name, sp_host, sp_domain, rna);
		if (sp_name[0] && sp_host[0]) {
			/* Haben wir ein oder mehrere '@' im name-Part?
			 * Das wuerde ZConnect-Software verwirren (die arme).
			 * Also nach % wandeln. Das wird in den meisten Teilen
			 * der Welt verstanden.
			 */
			while( (q=strchr(sp_name, '@')) != NULL )
				*q = '%';
			/* 
			 * smail and inn akzeptieren Adressen, die auf einen
			 * '.' enden. ZC-Software tut das nicht. smail beachtet
			 * den Punkt nicht, also koennen wir ihn auch
			 * wegwerfen.
			 */
			if(sp_domain[0])
				if(sp_domain[strlen(sp_domain)-1]=='.')
					sp_domain[strlen(sp_domain)-1]='\0';
			q=index(sp_name, '@');
				if(q) { *q='\0'; }
			q=index(sp_host, '@');
				if(q) { *q='\0'; }
			q=index(sp_domain, '@');
				if(q) { *q='\0'; }
			sprintf(zaddr, "%s@%s.%s", sp_name, sp_host, sp_domain);
			if (!sp_domain[0]) strcat(zaddr, "UUCP");
			if(rna[0] == '"')
			{
				for(q=rna+1; *q && *q!='"'; q++)
				{
					*(q-1)=*q;
				} 
				*(q-1)='\0';
			}
		/* Nach RFC822 darf der Lokalteil einer Adresse mit '"'
		   geklammert sein. qmail macht das auch generell so,
		   smail verwirft diese Klammern mindestens dann nicht,
		   wenn der Lokalteil auf '.' endet. Auf ZC-Seite ist
		   `"user."@host.do.main' aber etwas anderes als
		   `user.@host.do.main'. Sigh. Also verwerfen wir die
		   Hochkommata, wenn im Realnamen kein Leerzeichen auftritt. */
			if( '"' == sp_name[0] && NULL == index(sp_name, ' ') )
			{
				for(q=sp_name+1; *q && '"' != *q; q++)
					*(q-1)=*q;
				*(q-1)='\0';
			}
			if( count(zaddr, '@') == 1 )
			{
				if (rna[0])
				{
					char *s=decode_mime_header(rna);
#ifdef CONVERT_ISO
					iso2pc(s, strlen(s));
#endif
					fprintf(f, "%s: %s (%s)\r\n", zconnect_header, zaddr, s);
					dfree(s);
				}else{
					fprintf(f, "%s: %s\r\n", zconnect_header, zaddr);
				}
				counter++;
			} else {
				/* Na prima. Eine illegale Adresse, die keinen
				   oder mehrere @-Zeichen enthlt.
				   Wie ist die eigentlich bis hierhin vorgedrungen?
				   Wir knnen aber schlecht keinen entsprechenden
				   Header ausgeben...
				   Den rna lassen wir unter den Tisch fallen,
				   im Realnamenfeld darf keine weitere Klammer
				   auftauchen.
				 */
				fprintf(f, "%s: nobody@nowhere.universe (%s)\r\n",
				  zconnect_header, zaddr);
				counter++;
			}
		}
		 else if(sp_name[0]) {
		 /* Nur Username ohne @... . Absolut unbrauchbar,
		  * aber wir leiten hier ja nur weiter...
		  */
			if(rna[0])
				fprintf(f, "%s: %s (%s)\r\n", zconnect_header, sp_name, rna);
			else
				fprintf(f, "%s: %s\r\n", zconnect_header, sp_name);
			counter++;
		}
		if (!komma) break;
		if (max > 0 && counter >= max) break; 
		for (start = komma+1; *start && isspace(*start); start++)
			;
	}
	dfree(line);

	return counter;
}

char *printpath(char *reverse_path)
{
	char *p, *a, *rot;

	if (!reverse_path) {
		rot = dstrdup("");
		return(rot);
	}
	p = dstrdup(reverse_path);
	/* Der Routweg ist ebensolang wie die Adresse vorher,
	   schlielich entsteht er daraus durch Umsortieren.
	 */
	rot = dalloc(strlen(reverse_path)+1);
	rot[0] = '\0';
	a = fqdn_at(p);
	if (a) {
		*a = '\0';
		sprintf(rot, "%s!", a+1); /* rot ist gro genug dafuer. */
	}
	a = strrchr(p, '!');
	if (a) *a = '\0';
	strcat(rot, p);
	dfree(p);
	return(rot);
}

header_p convheader(header_p hd, FILE *f, char *host, char *from)
{
	extern int main_is_mail;
	int has_wab=0;
	header_p p, t;
	char *s, *st, *e, *to;
	static time_t mid_counter = 0;

	s = st = e = to = NULL;
	p = find(HD_X_GATEWAY, hd);
	if (p) {
		s = strstr(p->text, "NETCALL3.8");
		if (s) {
			while (s && *s && !isspace(*s))
				s++;
			while (s && *s && isspace(*s))
				s++;
			e = strchr(s, ',');
			if (e) *e = '\0';
			fprintf(f, HN_ZNETZ_CONV": %s\r\n", s);
		}
	}
	p = find(HD_UU_FOLLOWUP_TO, hd);
	if (p) {
		for (st = p->text; isspace(*st); st++)
			;
		if (stricmp(st, "poster") == 0) {
			p = find(HD_UU_SENDER, hd);
			if (!p) p = find(HD_UU_REPLY_TO, hd);
			if (!p) p = find(HD_UU_FROM, hd);
			if (p) convaddr(HN_DISKUSSION_IN, p->text, 0, f);
		} else
			printbretter(p->text, HN_DISKUSSION_IN, f);
		hd = del_header(HD_UU_FOLLOWUP_TO, hd);
	}
	p = find(HD_UU_NEWSGROUPS, hd);
	if (p) {
		printbretter(p->text, main_is_mail ? HN_KOP : HN_EMP, f);
		hd = del_header(HD_UU_NEWSGROUPS, hd);
	}
	p = find(HD_UU_MESSAGE_ID, hd);
	if (p) {
		fprintf(f, HN_MID": ");
		s=strchr(p->text, '<');
		if(s)
		{
			for (s++; *s && *s != '>'; s++)
				putc(*s, f);
		}else{
			fputs(p->text,f);
		}
		/* will niemals News-Message-IDs veraendern. Auch
		 * nicht, wenn sie keinen @ enthalten.
		 */
		s=strchr(p->text,'@');
		if(s==NULL && main_is_mail)
		{
			fputs("@unknown.nil",f);
		}
		fputs("\r\n", f);
		hd = del_header(HD_UU_MESSAGE_ID, hd);
	}
	else {
		/* Keine Message-ID */
		if (main_is_mail) {
			if (mid_counter == 0)
				mid_counter = time(NULL);
			fprintf(f, HN_MID": GATECONV.%d.%ld@%s.%s\r\n",
			(int)getpid(), (long)mid_counter++,
			boxstat.boxname, boxstat.boxdomain);
		}
	}
	p = find(HD_UU_FROM, hd);
	if (p) {
		if (convaddr(HN_ABS, p->text, 1, f) == 0) {
			/*
			 *  Illegale Nachricht - die sollte uns gar nicht
			 *  erst erreicht haben, aber weder C-News noch INN
			 *  testen sorgfltig.
			 *
			 *  Jetzt stehen wir ohne Absender da!
			 *
			 *  Heiko Schlichting schlgt vor, so eine Nachricht
			 *  einfach zu lschen, aber das ist (momentan) an
			 *  dieser Stelle nicht so einfach, daher setzen
			 *  wir erstmal einen Dummy-Absender ein.
			 */
			 if (!sp_name[0]) strcpy(sp_name, "nobody");
			 if (!sp_host[0]) strcpy(sp_host, "nowhere");
			 if (!sp_domain[0]) strcpy(sp_domain, "universe");
			 fprintf(f, HN_ABS": %s@%s.%s (Don't mail me!)\r\n", sp_name, sp_host,
		   		sp_domain); 
		}
		if (wab_name[0] && (stricmp(sp_name, wab_name) != 0 || stricmp(sp_host, wab_host) != 0
		   || stricmp(sp_domain, wab_domain) != 0)) {
			/* 
			 * WAB: der Form !User@system. darf nicht kommen.
			 * Einfach wegwerfen ist zwar keine gute Loesung,
			 * aber sonst kommen die Nachrichten gar nicht an.
			 *
			 * Diese WAB: entstehen bei Einlieferungen der Form
			 * @mail.other.do.main:USER@system
			 */
		 	if(wab_name[0] && (wab_name[0]!='!') && (wab_name[0]!='@')
		 		&& wab_host[0] && wab_domain[0])
		 	{
		   		fprintf(f, HN_WAB": %s@%s.%s\r\n", wab_name, wab_host,
		   			wab_domain);
			/* 
			 * Wir haben einen WAB: geschrieben, also sollte gleich kein
			 * zweiter kommen...
			 */
		 		has_wab=1;
		   		hd = del_header(HD_UU_SENDER, hd);
		   	}
		}
		hd = del_header(HD_UU_FROM, hd);
	} else {
		if (convaddr(HN_ABS, from, 1, f) == 0) {
			 if (!sp_name[0]) strcpy(sp_name, "nobody");
			 if (!sp_host[0]) strcpy(sp_host, "nowhere");
			 if (!sp_domain[0]) strcpy(sp_domain, "universe");
			 fprintf(f, HN_ABS": %s@%s.%s (Don't mail me!)\r\n", sp_name, sp_host,
		   		sp_domain); 
		}
	}
	p = find(HD_UU_SENDER, hd);
	if (p)
	{
	/*
	 * Sender: abc
	 * wandeln wir nicht in einen WAB:.
	 * Sender: mit zwei oder mehr @ auch nicht.
	 */
		if(count(p->text,'@')==1 && 0==has_wab)
		{
			convaddr(HN_WAB, p->text, 1, f);
			has_wab=1;
			hd = del_header(HD_UU_SENDER, hd);
		}
	} else { /* eigentlich auch, wenn if(count... fehlgeschlagen ist, */
		p = find(HD_UU_RESENT_FROM, hd);
		if(p && 0==has_wab)
		{
			if(count(p->text,'@')==1)
			{
				convaddr(HN_WAB, p->text, 1, f);
				has_wab=1;
				hd = del_header(HD_UU_RESENT_FROM,hd);
			}
		}
	}
	p = find(HD_UU_SUBJECT, hd);
	if (p) {
		char *s;
		s=decode_mime_header(p->text);
#ifdef CONVERT_ISO
		iso2pc(s, strlen(s));
#endif
		fprintf(f, HN_BET": %s\r\n", s);
		hd = del_header(HD_UU_SUBJECT, hd);
		dfree(s);
	} else
		fputs(HN_BET": -\r\n", f);
	p = find(HD_UU_DATE, hd);
	if (p) {
		int tz;
		char *datum;
		struct tm *lt;
		time_t dt;

		datum = dstrdup(p->text);
		dt = parsedate(datum, &tz);
		dfree(datum);
		if (dt == -1) {
			tz = 0;
			dt = time(NULL);
#ifdef LOG_ERRORS_IN_HEADERS
			fputs("U-X-Date-Error: parse error in Date: line\r\n",f);
#endif
		}
		fprintf(f, HN_UU_U_DATE": %s\r\n", p->text);
		dt += (3600 * (boxstat.timezone));
		lt = gmtime(&dt);
		fprintf(f, HN_EDA": %04d%02d%02d%02d%02d%02dW%+d:%02d\r\n",
			lt->tm_year+1900, lt->tm_mon+1, lt->tm_mday,
			lt->tm_hour, lt->tm_min, lt->tm_sec,
			tz/60, abs(tz%60));
		hd = del_header(HD_UU_DATE, hd);
	}
	p = find(HD_UU_IN_REPLY_TO, hd);
	if (p) {
		s = strchr(p->text, '<');
		if (s) s = strchr(s, '>');
		if (s) {
			fprintf(f, HN_BEZ": ");
			/* Hier kommen wir nur hin, wenn es in p->text
			   '<'...'>' gibt. */
			for (s=strchr(p->text, '<')+1; *s && *s != '>'; s++)
				putc(*s, f);
			fputs("\r\n", f);
			hd = del_header(HD_UU_IN_REPLY_TO, hd);
		}
	}
	p = find(HD_UU_CONTROL, hd);
	if (p) {
		char *brett;
		fprintf(f, HN_UU_U_CONTROL": %s\r\n", p->text);
		fputs(HN_STAT": CTL\r\n", f);
		if(strncasecmp(p->text,"cancel ",strlen("cancel "))==0) {
			s=strchr(p->text, '<');
			if (s) {
				char *mid;
				int istda;
				mid = dstrdup(s+1);
				s = mid + strcspn(mid, "> ");
				if(s)
					*s = '\0';
				fprintf(f, HN_CONTROL": CANCEL %s\r\n",mid);
				/* OK. Ist die MID auch bei den
				 * References dabei, oder mssen wir
				 * den BEZ noch zustzlich erzeugen? */
				istda = 0;
				p = find(HD_UU_REFERENCES, hd);
				if(p)
					for(s = p->text; *s; s += strcspn(s,"<")) {
						if(*s=='<') s++;
						if(strncmp(mid,s,strcspn(s,"> "))==0) {
							istda=1;
							break;
						}
					}
				if(0==istda) /* Wir muessen selber... */
				fprintf(f, HN_BEZ": %s\r\n",mid);
				dfree(mid);
			}
		}
		else if (strncasecmp(p->text,"newgroup ", strlen("newgroup "))==0) {
			fputs(HN_CONTROL": ADD ", f);
			brett = dstrdup(p->text + strlen("newgroup "));
			/* Flags (z.B. unmoderated) abschneiden */
			s = strchr( brett, ' ');
			if ( s != NULL)
				*s = '\0';
			printbretter(brett, NULL, f);
			dfree(brett);
		}
		else if (strncasecmp(p->text,"rmgroup ", strlen("rmgroup "))==0) {
			fputs(HN_CONTROL": DEL ", f);
			brett = dstrdup(p->text + strlen("rmgroup "));
			/* Flags (z.B. unmoderated) abschneiden */
			s = strchr( brett, ' ');
			if ( s != NULL)
				*s = '\0';
			printbretter(brett, NULL, f);
			dfree(brett);
		}
		hd = del_header(HD_UU_CONTROL, hd);
	}
	p = find(HD_UU_REFERENCES, hd);
	if (p) {
		char buffer[MAXLINE], *t;

		for (s=strchr(p->text, '<'); s && *s; s=strchr(s, '<')) {
			fputs(HN_BEZ": ", f);
			for (t = buffer, s++; *s && *s != '>'; s++) {
				putc(*s, f);
				*t++ = *s;
			}
			*t = '\0';
			fputs("\r\n", f);
		}
		hd = del_header(HD_UU_REFERENCES, hd);
	}
	p = find(HD_UU_ORIG_TO, hd);
	if (p) {
	/* xxx //// mehrfach? */
		convaddr(HN_OEM, p->text, 0, f);
		hd = del_header(HD_UU_ORIG_TO, hd);
	}
	p = find(HD_UU_TO, hd);
	if (p) {
#ifdef UUCP_SERVER
		fprintf(f, HN_UU_U_TO": %s\r\n", p->text);
#endif
		to = dstrdup(p->text);
		if (!main_is_mail)
			convaddr(HN_KOP, p->text, 0, f);
		hd = del_header(HD_UU_TO, hd);
	}
	p = find(HD_UU_CC, hd);
	if (p) {
#ifdef UUCP_SERVER
		fprintf(f, HN_UU_U_CC": %s\r\n", p->text);
#endif
		convaddr(HN_KOP, p->text, 0, f);
		hd = del_header(HD_UU_CC, hd);
	}
	p = find(HD_UU_BCC, hd);
	if (p) {
#ifdef UUCP_SERVER
		fprintf(f, HN_UU_U_BCC": %s\r\n", p->text);
#endif
		convaddr(HN_KOP, p->text, 0, f);
		hd = del_header(HD_UU_BCC, hd);
	}
/*
 * Inzwischen wird der Routepfad direkt durchgereicht, um Probleme
 * nach dem Rueckkonvertieren zu vermeiden. Das ist zwar auf ZC-Seite
 * strenggenommen falsch, fuehrt aber zu den wenigsten Problemen.

 * Andererseits kann man damit den MAPS im Zerberus nicht verwenden.
 * Also gibt es die Mglichkeit, Nachrichten an MAPS mit einem 
 * ZC-konformen Routstring zu schreiben.

 * Nachdem wir dafr aber meistens Nachrichten ohne Routstring (da vom
 * lokalen Host kommend) vorgesetzt bekommen werden, kopieren wir
 * den Routstring zunaechst in einen eigenen Speicherbereich und
 * arbeiten damit weiter.

 */
	{
		char *rot;

		rot=NULL;
		if (main_is_mail) {
			p = find(HD_UU_X_ZC_ROT, hd);
			if (p)
				rot = dstrdup(p->text);
		}
		else {
			p = find(HD_UU_PATH, hd);
			if (p)
				rot = dstrdup(p->text);
		}
		if (!rot)
				rot = printpath(from);

		hd = del_header(HD_UU_PATH, hd);
		hd = del_header(HD_UU_X_ZC_ROT, hd);

		fprintf(f, HN_ROT": %s\r\n", rot);
		dfree(rot);
	}
	p = find(HD_UU_RETURN_RECEIPT_TO, hd);
	if (p) {
		convaddr(HN_EB, p->text, 0, f);
		hd = del_header(HD_UU_RETURN_RECEIPT_TO, hd);
	}
	p = find(HD_UU_ORGANIZATION, hd);
	if (p) {
		char *s;
		s=decode_mime_header(p->text);
#ifdef CONVERT_ISO
		iso2pc(s, strlen(s));
#endif
		fprintf (f, HN_ORG": %s\r\n", s);
		hd = del_header(HD_UU_ORGANIZATION, hd);
		dfree(s);
	}
        p = find(HD_UU_SUPERSEDES, hd);
        if (p) {
        	/* Patch von Moritz - Null Pointer sonst moeglich */
               	s=strchr(p->text, '<');
               	if (s) {
               	        fprintf(f, HN_ERSETZT": ");
               	        for (s++; *s && *s != '>' && *s != ' '; s++)
               	                putc(*s, f);
               	        putc('\r', f); putc('\n', f);
               	}
		hd = del_header(HD_UU_SUPERSEDES, hd);
	}
	p = find(HD_UU_EXPIRES, hd);
	if (p) {
		struct tm *lt;
		int tz;
		char *datum;
		time_t dt;

		datum = dstrdup(p->text);
		dt = parsedate(datum, &tz);
		dfree(datum);
		if (dt == -1) {
			tz = 0;
			dt = time(NULL);
#ifdef UUCP_SERVER
			fputs ("U-X-Date-Error: parse error in Expires: line\r\n", f);
#endif
		}
		fprintf (f, HN_UU_U_EXPIRES": %s\r\n", p->text);
		dt += (3600 * (boxstat.timezone ));
		lt = gmtime(&dt);
		fprintf(f, HN_LDA": %04d%02d%02d%02d%02d%02dW%+2d:%02d\r\n",
			lt->tm_year+1900, lt->tm_mon+1, lt->tm_mday,
			lt->tm_hour, lt->tm_min, lt->tm_sec,
			tz/60, tz%60);
		hd = del_header(HD_UU_EXPIRES, hd);
	}
	p = find(HD_UU_REPLY_TO, hd);
	if (p) {
		convaddr(HN_ANTWORT_AN, p->text, 0, f);
		hd = del_header(HD_UU_REPLY_TO, hd);
	}
	p = find(HD_UU_X_MAILER, hd);
	if (p) {
		fprintf(f, HN_MAL": %s\r\n", p->text);
		hd = del_header(HD_UU_X_MAILER, hd);
	}
	p = find(HD_UU_LINES, hd);
	if (p) {
		fprintf(f, HN_UU_U_LINES": %s\r\n", p->text);
		hd = del_header(HD_UU_LINES, hd);
	}
	hd = del_header(HD_CONTENT_LENGTH, hd);

	p = find(HD_X_GATEWAY, hd);
	fprintf(f, HN_GATE": RFC1036/822 %s.%s [" 
				MAILER " " VERSION "#%u]%s%s\r\n",
				boxstat.boxname, boxstat.boxdomain, 
				cnumber, p ? ", " : "", p ? p->text : "");
	if (p)
		hd = del_header(HD_X_GATEWAY, hd);

	/* Fido Brettempfaenger */
	p = find(HD_UU_X_COMMENT_TO, hd);
	if (p) {
		fprintf(f, HN_F_TO": %s\r\n", p->text);
		hd = del_header(HD_UU_X_COMMENT_TO, hd);
	}
	
	/* eigentlich schner wre es, diese Header oben zuerst auszugeben
	   und dann entsprechende Wandlungen zu unterbinden. */

	for (p = hd; p; p = t) {
		t = p->next;
		if (  strncasecmp(p->header, "X-ZC-", 5) == 0
		   || strncasecmp(p->header, "X-Z-", 4) == 0 ) {
			s=index(index(p->header,'-'),'-')+1;
			if ( strcasecmp(s, "WAB") == 0
			  || strcasecmp(s, "ABS") == 0
			  || strcasecmp(s, "MID") == 0 ) {
				hd=del_header(p->code,hd);
			} 
		}
	}
	
	/* Gib X-ZC- und X-Z- und X-FTN- Headerzeilen aus.
	 * Wandele sie, falls sie kodierte 8-Bit-Zeichen haben.
	 */
	for (p = hd; p; p = t) {
		char *s;
		int ip = 0;
		char *praefix = "";

		t = p->next;
		if (strncasecmp(p->header, "X-ZC-", 5) == 0)
			ip = 5;
		else if (strncasecmp(p->header, "X-Z-", 4) == 0)
			ip = 4;
		else if (strncasecmp(p->header, "X-FTN-", 6) == 0) {
			ip = 6;
			praefix = "F-";
		}
		
		if (ip) {
			header_p t1;

			for (t1 = p; t1; t1 = t1->other) {
				s = decode_mime_header(t1->text);
				iso2pc(s, strlen(s));
				fprintf(f, "%s%s: %s\r\n", praefix, 
						(t1->header) + ip, s);
				dfree(s);
			} 
			hd = del_header(p->code, hd);
		}
	}

	return hd;
}


#ifdef CONVERT_ISO
/* Gibt !=0 zurueck, wenn in den ersten Bytes von buf ueber
 * 50 % Zeichen sind, die Bit 7 gesetzt haben (Loesung, um
 * Irrtuemer beim Zeichensatz herauszufiltern, z.B. russische
 * Fonts u. dgl., die duch Konvertierung zerstoert wuerden)
 *
 * Suche in den ersten 512 Bytes reicht erfahrungsgem nicht.
 * Z.B. kann ein ASCII-Mailheader vor dem Text sein.
 * Wir versuchen mal 10 KByte, das sollte reichen.
 */
int over50_8bit(char *buf, size_t len) {
	int li, i, b8;

	i = li = (int)(len > (size_t)10240 ? 10240 : len);
	b8 = 0;

	while(i--)
		if (buf[i] & 0x80)
			b8++;
	
	return b8 > li/2;
}
#endif


/* Nachrichteninhalt und letzte Headerzeilen ausgeben je nach Typ und
 * Codierung
 */

int make_body(char *bigbuffer, size_t msglen, 
		mime_header_info_struct *mime_info,
		int binaer, char *smallbuffer, FILE *zconnect) {
	int eightbit;
#ifdef APC_A2B
	size_t latob;
#endif

#ifdef APC_A2B
/* APC-typische Binaer-Nachrichten dekodieren. Wenn wir eine
 * solche haben, behandeln wir sie ganz separat von allem anderen.
 */
	latob = apc_a2b(bigbuffer, &msglen);
	if (latob > -1) {
		char *start = bigbuffer;
		fprintf(zconnect, HN_TYPE": BIN\r\n");
		/* ein CR/LF nehmen wir weg, es gehoeht noch zum
			RFC-Header */
		if (latob > 0 && *start == '\r') {
			start++; msglen--; latob--;
		}
		if (latob > 0 && *start == '\n') {
			start++; msglen--; latob--;
		}
		if (latob > 0) {
			/* Text, der der Nachricht vorangestellt ist */
#ifdef CONVERT_ISO
			iso2pc(start, latob);
#endif
			fprintf(zconnect, HN_KOM": %ld\r\n", latob);
		}
		fprintf(zconnect, HN_LEN": %ld\r\n\r\n", msglen);
		fwrite(start, (size_t) msglen, 1, zconnect);
	}
	else
#endif /* APC_A2B */

	/* Hier kommt die UnixConnect-Spezial-Multipart-Abteilung.
	 * Es handelt sich um ehemalige binaere ZConnect-Nachrichten, die
	 * einen KOM:-Header hatten.
	 * Wir machen das nur, wenn sie oben nicht MIME-Dekodiert wurde,
	 * damit wir sicher sind, dass wir am Ende der Nachricht (und
	 * erst dort) ein Nullbyte haben. Dann koennen wir strchr()
	 * und dergleichen benutzen.
	 */
	if (!binaer && mime_info->unixconnect_multipart) {
		char *p1, *p2;
		size_t wrlen;

		wrlen=msglen;
		/* Erste border nach p1 */
		p1 = strstr(bigbuffer, "--"SP_MULTIPART_BOUNDARY);
		if (p1)
			/* LF vor Border zw. KOM: und Binaerdaten nach p2 */
			p2 = strstr(p1+5, "\n--"SP_MULTIPART_BOUNDARY);
		else
			p2 = NULL;
		if (p1 && p2) {
			int i;

			/* boundary, maximal 2 Headerzeilen 
			 * und Leerzeile ueberspringen
			 */
			for (i=4; i>0; i--) {
				while (*p1 && *p1 != '\n')
					p1++;
				p1++;
				if ((*p1 == '\r' || *p1 == '\n') 
					&& i>2)
					i = 2; /* Leerzeile noch ueberspringen */
			}
			/* KOM: Text ist jetzt p1 bis p2 (incl.!) */
			msglen = p2-p1+1;
			memcpy(smallbuffer, p1, msglen);
			/* CHARSET gilt fuer KOM:-Teil nicht */
#ifdef CONVERT_ISO
			iso2pc(smallbuffer, msglen);
#endif
			fprintf (zconnect, HN_KOM": %ld\r\n", (long)msglen);

			/* Ende der Boaderzeile */
			while (*p2 && *p2 != '\n')
				p2++;
			p2++;
			/* Content-Type */
			while (*p2 && *p2 != '\n')
				p2++;
			p2++;
			p1 = strstr(p2, "--"SP_MULTIPART_BOUNDARY);
			if (p1)
				wrlen = p1 - p2;
			else
				wrlen -= (p2 - bigbuffer);
			/* Na, dann versuchen wir mal zu decodieren */
			if (decode_x_uuencode(p2, (long *)&wrlen, &eightbit, mime_info))
				fprintf(zconnect, HN_TYPE": BIN\r\n");
		} 
		else 
			wrlen = 0;
		if (mime_info->filename)
			fprintf(zconnect, HN_FIL": %s\r\n", mime_info->filename);
		fprintf(zconnect, HN_LEN": %ld\r\n\r\n", (long)(msglen+wrlen));
		fwrite(smallbuffer, msglen, 1, zconnect);
		fwrite(p2, wrlen, 1, zconnect);
	}
	else
	if (!binaer) {
		char *start;

		start = bigbuffer;

		/* Zeichensatz-Konvertierung des Body nur, wenn
		 * es sich um text/plain handelt (default)
		 * (ist auch bei nicht-MIME-Nachrichten gesetzt)
		 */
		if (mime_info->text_plain) {
#ifdef CONVERT_ISO
			/* Wenn Zeichensatz us-ascii, undefiniert oder iso-8859-1
			 * UND kein excessiver Gebrauch von des 8. Bit
			 * konvertieren wir nach IBM */

			if (mime_info->charset > 1 )
				fprintf(zconnect, HN_CHARSET": ISO%d\r\n", mime_info->charset);
			else
				if (!over50_8bit(start, (size_t) msglen))
					iso2pc(start, msglen);
				else {
					/* Auf Konvertierung verzichtet: Es ist u.U. eine
					 * Nachricht mit kyrillischer Schrift oder so
					 *
					 * Damit garantiert keine Zeichensatz-Wandlungen im
					 * ZConnect-Bereich stattfinden, machen wir eine
					 * Binaernachricht daraus :-(
					 */
					fprintf(zconnect, HN_TYPE": BIN\r\n"
						"X-Warning: Charset unknown; type binary to avoid charset conversion\r\n");
				}
#else /* CONVERT_ISO */
			/* Wir konvertieren nie nach IBM, setzen aber eine CHARSET:
			 * Headerzeile (default: ISO1)
			 */
			fprintf(zconnect, HN_CHARSET": ISO%d\r\n", 
				mime_info->charset >0 ? mime_info->charset : 1);
#endif /* CONVERT_ISO */
		}
		fprintf (zconnect, HN_LEN": %ld\r\n\r\n", (long)msglen);

		/* Inhalt ausgeben: */
		fwrite(start, (size_t)msglen, 1, zconnect);
	} else {
		/* Binaer */
		fprintf(zconnect, HN_TYPE": BIN\r\n");
		if (mime_info->filename) {
			fprintf(zconnect, HN_FIL": %s\r\n", mime_info->filename);
		}
		fprintf(zconnect, HN_LEN": %ld\r\n\r\n", (long)msglen);
		fwrite (bigbuffer, 1, msglen, zconnect);
	}
	return 0;
}

/* Konvertiere eine MIME-codierte Headerzeile nach
 * 8-bit.
 * "Q"-codierte Teile werden hier lokal behandelt.
 * "B"-codierte Teile sind genau base64 - dafr nehmen wir die
 * Funktion, die das auch im Body tut
 */
#define HEXCHAR2DEC(x)   (((x)&0xf) + ((x)>'9' ? 9 : 0))
static char *decode_mime_header(register char *s) {
	char *dec, *cp, *cpso, *cp2, *cp3;
	int encoded, just_off;
	char encoding;
	long len;
	int eightbit;

	encoded=0;
	dec=dalloc(strlen(s)+1);
	cp=dec;
	cpso=s;
	just_off=0;
	for(; *s; s++) {
		if (just_off == 1)
			just_off++;
		else
			just_off = 0;
			
		if (!encoded && *s == '=' && s[1] == '?') {
			cp2=NULL;
			/* iso-8859-1 ? */
			if (strncasecmp(s+2, iso_8859_1, sizeof(iso_8859_1)-1)
					== 0) {
				cp2 = s + 2 + sizeof(iso_8859_1)-1;
			}
			else if (strncasecmp(s+2, us_ascii, sizeof(us_ascii)-1)
					== 0) {
				cp2 = s + 2 + sizeof(us_ascii)-1;
			}

			if (cp2 && *cp2=='?' && cp2[1] && cp2[2]=='?') {
				encoding = toupper(cp2[1]);
				cp2 += 2;
				if (encoding=='Q') {
					encoded=1;
					s = cp2;
					continue;
				}
				else
				if (encoding=='B') {
					/* base64 jetzt sofort per
					 * Funktion decodieren.
					 * Das Ende des codierten Teils
					 * erkennen wir am '?'.
					 */
					cp2++;
					if ((cp3=strchr(cp2, '?')) != NULL
							&& cp3[1]=='=') {
						len = cp3-cp2;
						memcpy(cp, cp2, len);
						if (decode_base64(cp, &len,
								&eightbit)==0) {
							s=cp3 + 1;
							cp+= (int)len;
							continue;
						}
					}

				}
			}
		}
		if (encoded && *s == '?' && s[1] == '=') {
			encoded=0;
			s++;
			just_off=1;
			continue;
		}

		if (encoded && *s == '=' && isxdigit(s[1]) && isxdigit(s[2])) {
			*cp++ = (HEXCHAR2DEC(s[1])<<4) | HEXCHAR2DEC(s[2]);
			s+=2;
		}
		else if (encoded && *s == '_')
			*cp++ = (char)0x20;
		else if (just_off == 2 && !encoded && *s == '\t')
			/* we are right behind a '?=' and have a tab
			 * we discard it in favour of further encoded
			 * text
			 */
			;
		else
			*cp++ = *s;
	}
	*cp='\0';
	/* if encoded is true, the encoded string is incomplete. */
	if (encoded) 
		strcpy(dec, cpso);
	return dec;
}
